//Copyright Monogram (Grant & Union) 2021
#include "TCPMonogramThread.h"
#include "MonogramCore.h"

#include "Containers/UnrealString.h"
#include "Misc/CString.h"
#include "Sockets.h"
#include "Interfaces/IPv4/IPv4SubnetMask.h"
#include "Interfaces/IPv4/IPv4Address.h"
#include "Interfaces/IPv4/IPv4Endpoint.h"
#include "Common/TcpSocketBuilder.h"

const size_t TCPMonogramThread::BUFFER_SIZE = 100;

TCPMonogramThread::TCPMonogramThread(UMonogramCore* parentBlueprint) : threadStop(0)
{
	this->parentClass = parentBlueprint;
}

bool TCPMonogramThread::Init()
{
	return true;
}

void TCPMonogramThread::Stop()
{
	threadStop.Increment();
	return;
}

uint32 TCPMonogramThread::Run()
{
	//UE_LOG(LogTemp, Display, TEXT("Monogram Connection Thread started"));
	FSocket* Listener;
	FSocket* Connection;

	int32 bytesRead = 0;
	FIPv4Address localHost(127, 0, 0, 1);
	Listener = FTcpSocketBuilder(TEXT("MonogramSocket"))
		.AsReusable().AsBlocking().BoundToAddress(localHost).BoundToPort(47370).Listening(2).Build();

	if (!Listener)
	{
		//UE_LOG(LogTemp, Fatal, TEXT("Monogram: <Fail> Couldn't create listener socket."));
		return 99;
	}

	while (true)
	{
		uint8 dataBuffer[TCPMonogramThread::BUFFER_SIZE] = { 0 };
		bytesRead = 0;
		bool hasPending = false;
		//UE_LOG(LogTemp, Display, TEXT("Monogram: Waiting for Connection!"));
		Listener->WaitForPendingConnection(hasPending, FTimespan::FromSeconds(6));

		if (hasPending == true)
		{

			//UE_LOG(LogTemp, Display, TEXT("Monogram: Connected to Turaco"));
			Connection = Listener->Accept(TEXT("Monogram: accepting incoming connection"));

			if (!Connection)
			{
				//UE_LOG(LogTemp, Warning, TEXT("Monogram: <Fail> Couldn't accept Connection."));
				break;
			}

			FString carryOver = "";
			FString message = "";
			FString slashN = "\n";
			FString appended = "";
			while (threadStop.GetValue() == 0)
			{
				//if we have carry over and it contains a whole line just process that for now
				//otherwise read the socket
				if (carryOver.Contains("\n"))
				{
					bytesRead = 0;
					appended = carryOver;
				}
				else if (!Connection->Recv(dataBuffer, TCPMonogramThread::BUFFER_SIZE, bytesRead, ESocketReceiveFlags::None))
				{
					//UE_LOG(LogTemp, Warning, TEXT("Monogram: Connection to turaco lost, resetting socket."));
					break;
				}

				//if we have no data read skip this and assume we're just processing carryover
				if (bytesRead > 0)
				{
					int32 inCount = bytesRead;
					FString recvString(inCount, (TCHAR*)ANSI_TO_TCHAR((ANSICHAR*)dataBuffer));
					appended = carryOver.Append(recvString);
				}

				if (appended.Contains("\n"))
				{
					if (!appended.Split(slashN, &message, &carryOver))
					{
						UE_LOG(LogTemp, Warning, TEXT("Monogram: Nothing to split, this shouldn't happen."), *message);
						continue;
					}
				}
				else
				{
					//this should only happen if we didn't read *enough* so we don't have a line yet
					continue;
				}

				// This RegEx parses the simple protocol type messages
				const FRegexPattern MESSAGE_PATTERN(TEXT("^([a-z,A-Z,_,::]+)(?:(\\+|=)([^\r]*)\r?)?$"));
				FRegexMatcher matcher(MESSAGE_PATTERN, message);
				if (matcher.FindNext())
				{
					FString left, right; //maybe we just make it like __? double underscore?
					if (!matcher.GetCaptureGroup(1).Split("__", &left, &right)) {
						left = matcher.GetCaptureGroup(1);
					}
					FName action = FName(*left);
					FName command = FName(*right);
					UMonogramAction::Operation op = UMonogramAction::ParseOperation(matcher.GetCaptureGroup(2));
					FString opString = matcher.GetCaptureGroup(2);
					FName opName = FName(*opString);
					FName messageOut = FName(*message);
					float value = FCString::Atof(*matcher.GetCaptureGroup(3));

					TArray<UMonogramAction*> handlers;
					this->parentClass->MonogramActions.MultiFind(action, handlers, true);
					for (auto handler : handlers) {
						handler->Run(command, op, value);
					}
					//UE_LOG(LogTemp, Warning, TEXT("MONOGRAM: %s"), *message);
					//I'm not sure if it's an issue to call blueprintRunable functions from another thread.
					//So to be safe we'll post it to the main thread instead.
					AsyncTask(ENamedThreads::GameThread, [&]()
					{
						//this is running (with a reference capture on message, which doesn't exist if this function exits).
						if (parentClass->MonogramDataReceived.IsBound())
						{
							parentClass->MonogramDataReceived.Broadcast(messageOut, opName, action, value);
						}
					});
				}
				else
				{
					//UE_LOG(LogTemp, Warning, TEXT("Failed to parse message %s"), *message);
				}
			}
			Connection->Close();
			delete Connection;
		}
	}
	//UE_LOG(LogTemp, Display, TEXT("Exiting."));
	delete Listener;
	return 0;
}
